// Copyright (c) 2006 Advanced Micro Devices ("AMD").
//    
// All Rights Reserved.  Unpublished rights reserved under the 
// copyright laws of the United States of America, other countries 
// and international treaties.  The software is provided without 
// fee.  Permission to use, copy, store, modify, disclose, transmit 
// or distribute the software is granted, provided that this 
// copyright notice must appear in any copy, modification, 
// disclosure, transmission or distribution of the software.
//    
// AMD retains all ownership, copyright, trade secret and 
// proprietary rights in the software. THIS SOFTWARE HAS BEEN 
// PROVIDED "AS IS," WITHOUT EXPRESS OR IMPLIED WARRANTY INCLUDING, 
// WITHOUT LIMITATION, IMPLIED WARRANTIES OF MERCHANTABILITY, 
// FITNESS FOR A PARTICULAR USE AND NON-INFRINGEMENT.
//
// AMD does not assume or authorize any other person to assume for 
// it any liability in connection with the Software. AMD SHALL NOT 
// BE LIABLE TO COMPANY, OR ANY THIRD PARTY, IN CONTRACT, TORT, 
// WARRANTY, STRICT LIABILITY, OR OTHERWISE FOR ANY DIRECT DAMAGES, 
// OR FOR ANY SPECIAL, INDIRECT, INCIDENTAL OR CONSEQUENTIAL 
// DAMAGES, INCLUDING BUT NOT LIMITED TO, BUSINESS INTERRUPTION, 
// LOST PROFITS OR GOODWILL, OR LOSS OF INFORMATION EVEN IF AMD IS 
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

#include <windows.h>

// GeodeLX Display Registry structure
// This structure holds all user-modifiable registry
// information, such as panel enable, custom timings, etc.  This
// structure definition is taken from lxshared.h in the display
// driver source.  It is duplicated here to allow this code to be
// compilable and functional standalone.
//
typedef struct tagLXRegistrySettings
{
    unsigned long secretWord;           // Identifier to indicate valid buffer
    unsigned long unused;               // Unused DWORD
    unsigned long BIOSDetectOutput;     // Use the BIOS to detect the output device.
    unsigned long Compression;          // Compression Settings (enabled, auto, etc.)
    unsigned long FBLinear;             // Linear frame buffer setting
    unsigned long DDCEnabled;           // Enable DDC communication to a monitor
    unsigned long DDC_GPIOData;         // GPIO for DDC data line.
    unsigned long DDC_GPIOClock;        // GPIO for DDC clock line.
    unsigned long PanelEnable;          // Dynamic state of panel enable.
    unsigned long PanelWidth;           // Manually specified panel width for supported panel sizes.
    unsigned long PanelHeight;          // Manually specified panel height for supported panel sizes.
    unsigned long XVGAPanel;            // XVGA Panel for supported panel sizes.
    unsigned long SimultaneousCRT;      // Dynamic flag enabling simultaneous CRT display.    
    unsigned long ManualTimingEnable;   // Custom mode timings have been specified.    
    unsigned long EnablePanelScale;     // Allow modes to be stretched to fit the display.
    unsigned long EnableNon60HzPanel;   // Allow non-60Hz CRT modes when displaying on a TFT.
    unsigned long CustomPanelTimings;   // Custom panel timings have been specified.
    unsigned long RotationAngle;        // Angle of desktop rotation.
    unsigned long PanelRefresh;         // Detected refresh rate of the panel
    unsigned long PanelInSystem;        // Panel detected with BIOSDetectOutput or panel
                                        // specifically specified at install time.

} LX_REGISTRY_SETTINGS;

// Driver Escape Definitions
// These definitions are also pulled from lxshared.h in the display driver
// source.
//
#define GEODELX_GET_REGISTRY_SETTINGS  0x400
#define GEODELX_SET_REGISTRY_SETTINGS  0x401
#define GEODELX_GET_LAST_MODE          0x404
#define GEODELX_SECRET_WORD            0xABADD00D

// Output Type Definitions
//
#define GEODELX_OUTPUT_CRT   0
#define GEODELX_OUTPUT_TFT   1
#define GEODELX_OUTPUT_BOTH  2

// Global Variables
//
LX_REGISTRY_SETTINGS g_regData, g_oldSettings;
DEVMODE g_curMode;
DEVMODE g_newMode;

/////////////////////////////////////////////////////////////////////////////
// TFT_CRT
//
// Routine to set the output format to CRT, TFT or simultaneous.  This routine,
// in addition to programming the registry settings, attempts to set the last
// mode programmed for that display output.
//
int TFT_CRT (int path)
{
    ULONG lastMode[6];
    ULONG width, height;
    LONG modeset;
    BOOL retval;
    HDC dc;
    int mode = 0;
    DEVMODE dm;

    // Check for valid selection
    //
    if (path != GEODELX_OUTPUT_CRT && path != GEODELX_OUTPUT_TFT && 
        path != GEODELX_OUTPUT_BOTH)
    {
        return -1;
    }

    // Before we do anything, save a copy of the existing driver registry 
    // settings, such that we can restore them if the modeset fails.
    //
    dc = GetDC(NULL);
    g_oldSettings.secretWord = GEODELX_SECRET_WORD;
    retval = ExtEscape (dc, GEODELX_GET_REGISTRY_SETTINGS, sizeof(g_oldSettings), 
        (LPCSTR)&g_oldSettings, sizeof(g_oldSettings), (LPSTR)&g_oldSettings);
    ReleaseDC (NULL, dc);
    g_regData = g_oldSettings;

    // Verify valid communication with the driver.
    //
    if (retval <= 0)
        return -2;

    // Read the current display mode, in case we need to restore it.
    //
    EnumDisplaySettings (NULL, ENUM_CURRENT_SETTINGS, &g_curMode);
    g_newMode = g_curMode;

    // Update the registry settings.  This will potentially alter the list
    // of available modes.
    //
    switch (path)
    {
        case GEODELX_OUTPUT_CRT: 
            g_regData.PanelEnable = 0;
            break;
        case GEODELX_OUTPUT_TFT: 
            g_regData.PanelEnable = 1;
            g_regData.SimultaneousCRT = 0;
            break;
        default:
            g_regData.PanelEnable = 1;
            g_regData.SimultaneousCRT = 1;
            break;
    }            
    dc = GetDC(NULL);
    g_regData.secretWord = GEODELX_SECRET_WORD;
    retval = ExtEscape (dc, GEODELX_SET_REGISTRY_SETTINGS, sizeof(g_regData), 
        (LPCSTR)&g_regData, sizeof(g_regData), (LPSTR)&g_regData);
    
    if (retval <= 0)
        return -3;

    // Read the last programmed mode for this output format from the registry
    //
    lastMode[0] = GEODELX_SECRET_WORD;
    retval = ExtEscape (dc, GEODELX_GET_LAST_MODE, sizeof(lastMode), (LPCSTR)&lastMode[0], 
        sizeof(lastMode), (LPSTR)&lastMode[0]);
    ReleaseDC (NULL, dc);

    // Values returned are as follows:
    // lastMode[2] = last set width
    // lastMode[3] = last set height
    // lastMode[4] = last set refresh
    // lastMode[5] = last set rotation
    //
    if (retval)
    {
        // Extract the unrotated dimensions of the last mode.
        //
        if (lastMode[5] == 90 || lastMode[5] == 270)
        {
            width  = lastMode[3];
            height = lastMode[2];
        }
        else
        {
            width = lastMode[2];
            height = lastMode[3];
        }

        // Apply the current rotation
        //
        if (g_regData.RotationAngle == 90 || g_regData.RotationAngle == 270)
        {
            unsigned long temp = width;
            width = height;
            height = temp;
        }

        g_newMode.dmPelsWidth = width;
        g_newMode.dmPelsHeight = height;
        g_newMode.dmDisplayFrequency = lastMode[4];
    }
    else
        return -4;

    // Enumerate all display modes.  This forces calls to DrvGetModes, which allows the call 
    // to set the rotated mode to work.
    //    
    mode = 0;
    while (EnumDisplaySettings (NULL, mode, &dm))
        mode++;

    // Change the hardware display mode.  A modeset is needed to actually make the
    // output change take effect.
    //
    modeset = ChangeDisplaySettings (&g_newMode, CDS_UPDATEREGISTRY | CDS_GLOBAL);
    modeset = ChangeDisplaySettings (&g_newMode, CDS_RESET);

    if (modeset != DISP_CHANGE_SUCCESSFUL)
    {
        // Try again with default values.  One possible scenario here is that
        // the framebuffer size has changed or that the last set TFT mode was
        // a fake mode that was set coming out of a resume.
        //
        if (g_regData.PanelEnable)
        {
            g_newMode.dmPelsWidth  = g_regData.PanelWidth;
            g_newMode.dmPelsHeight = g_regData.PanelHeight;
            g_newMode.dmDisplayFrequency = g_regData.PanelRefresh;
        }
        else
        {
            g_newMode.dmPelsWidth  = 800;
            g_newMode.dmPelsHeight = 600;
            g_newMode.dmDisplayFrequency = 60;
        }

        // Adjust for rotation
        //
        if (g_regData.RotationAngle == 90 ||
            g_regData.RotationAngle == 270)
        {
            long temp = g_newMode.dmPelsWidth;
            g_newMode.dmPelsWidth = g_newMode.dmPelsHeight;
            g_newMode.dmPelsHeight = temp;
        }

        modeset = ChangeDisplaySettings (&g_newMode, CDS_UPDATEREGISTRY | CDS_GLOBAL);
        modeset = ChangeDisplaySettings (&g_newMode, CDS_RESET);
    }

    if (modeset != DISP_CHANGE_SUCCESSFUL)
    {
        // Put things back the way they were (or try to).  This implies that we 
        // revert the driver registry settings and change the mode registry settings.
        //
        dc = GetDC(NULL);
        retval = ExtEscape (dc, GEODELX_SET_REGISTRY_SETTINGS, sizeof(g_oldSettings), 
            (LPCSTR)&g_oldSettings, sizeof(g_oldSettings), (LPSTR)&g_oldSettings);
        ReleaseDC (NULL, dc);

        mode = 0;
        while (EnumDisplaySettings (NULL, mode, &dm))
            mode++;

        ChangeDisplaySettings (&g_curMode, CDS_UPDATEREGISTRY | CDS_GLOBAL);
        modeset = ChangeDisplaySettings (&g_curMode, CDS_RESET);
        return -5;
    }

    // 0 indicates success
    //
    return 0;
}
